@@ -17,10 +17,15 @@ import pandas
 import pandera
 import logging
 import sys
+import requests
+from requests.adapters import HTTPAdapter
+
+from requests.sessions import session
+from urllib3.util.retry import Retry
 from .._utils._constants import WORK_ORDER_STATUS, WORK_ORDER_CATEGORY, WORK_ORDER_PRIORITY
 
 __all__ = ['generate_password', 'convert_to_sqm', 'convert_to_pascal_case', 'ValidatedSetterProperty', 'ApiInputs',
-           'DiscoveryIntegrationInput', 'DataFeedFileProcessOutput']
+           'DiscoveryIntegrationInput', 'DataFeedFileProcessOutput', 'requests_retry_session']
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.DEBUG)
@@ -86,8 +91,8 @@ ApiInputs = namedtuple('ApiInputs',
                         'bearer_token', 'api_base_url', 'api_projects_endpoint', 'subscription_key', 'api_headers'])
 
 DiscoveryIntegrationInput = namedtuple(
-        'DiscoveryIntegrationInput',
-        ['api_project_id', 'installation_id', 'network_device_id', 'integration_device_id', 'user_id', 'batch_no'])
+    'DiscoveryIntegrationInput',
+    ['api_project_id', 'installation_id', 'network_device_id', 'integration_device_id', 'user_id', 'batch_no'])
 # DiscoveryIntegrationInput.__doc__ = """DiscoveryIntegrationInput(api_project_id, installation_id, network_device_id, integration_device_id, user_id, batch_no)
 #
 # Defines the required inputs to be provided to the run_discovery() method's integration_input parameter.
@@ -202,6 +207,46 @@ def convert_to_pascal_case(text: Union[str, List[str]]):
         return False
 
 
+def requests_retry_session(retries=5, status_forcelist=[429, 500, 502, 503, 504], backoff_factor=0.3, session=None, method_whitelist=["GET"]):
+    """
+    Provide retry policy for requests fired.
+
+    Parameters
+    ----------
+    retries : number
+        Number of retries the retry would do before declaring it as an exception.
+        Defaults to 5.
+    status_forcelist : List[number]
+        A set of integer HTTP status codes that we should force a retry on.
+        Defaults to 429, 500, 502, 503, 504.
+    backoff_factor : number
+        A backoff factor to apply between attempts after the second try (most errors are resolved immediately by a second try without a delay).
+        {backoff factor} * (2 ** ({number of total retries} - 1)) seconds
+        Defaults to 0.3
+    method_whitelist : List[string]
+        Set of uppercased HTTP method verbs that we should retry on
+        Defaults to: GET
+
+    Returns
+    -------
+        Session containing the retry configuration.
+        Can be changed as requests_retry_session().get() or .post() depending on requirements
+    """
+    session = session or requests.Session()
+    retry = Retry(
+        total=retries,
+        read=retries,
+        connect=retries,
+        status_forcelist=status_forcelist,
+        backoff_factor=backoff_factor,
+        method_whitelist=method_whitelist
+    )
+    adapter = HTTPAdapter(max_retries=retry)
+    session.mount('http://', adapter=adapter)
+    session.mount('https://', adapter=adapter)
+    return session
+
+
 _site_schema = pandera.DataFrameSchema(
     {
         'InstallationName': pandera.Column(pandera.String, checks=pandera.Check.str_length(1, 255)),
