@@ -6,6 +6,7 @@
 """
 A module for integrating asset creation, asset updates, data ingestion, etc into the Switch Automation Platform.
 """
+import json
 import re
 import sys
 import pandas
@@ -13,12 +14,13 @@ import pandas as pd
 import pandera
 import requests
 import datetime
+import pytz
 import logging
 import uuid
 from .._utils._platform import _get_structure, Blob
 from .._utils._utils import (ApiInputs, _with_func_attrs, _column_name_cap, _work_order_schema,
                              DiscoveryIntegrationInput, convert_to_pascal_case)
-from ..integration._utils import _timezone_offsets, _upsert_entities_affected_count, _adx_support
+from ..integration._utils import _timezone_offsets, _upsert_entities_affected_count, _adx_support, _timezone_dst_offsets
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.DEBUG)
@@ -32,7 +34,7 @@ consoleHandler.setFormatter(formatter)
 
 
 @_with_func_attrs(df_required_columns=['ApiProjectId', 'InstallationId', 'NetworkDeviceId', 'UserId', 'BatchNo',
-                                       'DriverUniqueId', 'Timestamp', 'DiscoveredValue'],
+                                       'DriverUniqueId', 'Timestamp', 'DiscoveredValue', 'DriverClassName'],
                   df_optional_columns=['DriverDeviceType', 'ObjectPropertyTemplateName', 'UnitOfMeasureAbbrev',
                                        'DeviceName', 'DisplayName', 'EquipmentType', 'EquipmentLabel', ])
 def upsert_discovered_records(df: pandas.DataFrame, api_inputs: ApiInputs, discovery_properties_columns: list,
@@ -72,7 +74,7 @@ def upsert_discovered_records(df: pandas.DataFrame, api_inputs: ApiInputs, disco
     data_frame = df.copy()
 
     required_columns = ['ApiProjectId', 'InstallationId', 'NetworkDeviceId', 'UserId', 'BatchNo', 'DriverUniqueId',
-                        'Timestamp', 'DiscoveredValue']
+                        'Timestamp', 'DiscoveredValue', 'DriverClassName']
     optional_columns = ['DriverDeviceType', 'ObjectPropertyTemplateName', 'UnitOfMeasureAbbrev', 'DeviceName',
                         'DisplayName', 'EquipmentType', 'EquipmentLabel', ]
     proposed_columns = data_frame.columns.tolist()
@@ -89,7 +91,7 @@ def upsert_discovered_records(df: pandas.DataFrame, api_inputs: ApiInputs, disco
     if discovery_properties_columns is not None and not set(discovery_properties_columns).issubset(data_frame.columns):
         logger.exception('Missing expected discovery property column(s): %s',
                          set(discovery_properties_columns).difference(proposed_columns))
-        return 'Integration.upsert_discovered_records(): data_frame must contain the following tag columns: ' + \
+        return 'Integration.upsert_discovered_records(): data_frame must contain the following discovery property column(s): ' + \
                ', '.join(discovery_properties_columns)
     elif discovery_properties_columns is None:
         logger.exception('Missing expected discovery property column(s): %s',
@@ -156,7 +158,7 @@ def upsert_discovered_records(df: pandas.DataFrame, api_inputs: ApiInputs, disco
 
     data_frame = set_properties(raw_df=data_frame, column_dict=column_dict)
     data_frame = data_frame.drop(columns=discovery_properties_columns + ['Timestamp'] + device_tag_columns +
-                                         sensor_tag_columns + metadata_columns  # + ontology_tag_columns
+                                 sensor_tag_columns + metadata_columns  # + ontology_tag_columns
                                  )
     data_frame = data_frame.assign(OntologyTags=None)
 
@@ -450,7 +452,7 @@ def upsert_sites(df: pandas.DataFrame, api_inputs: ApiInputs, tag_columns: list
 
     if tag_columns is not None and not set(tag_columns).issubset(data_frame.columns):
         logger.exception('Missing expected tag column(s): %s', set(tag_columns).difference(proposed_columns))
-        return 'Integration.upsert_device_sensors(): data_frame must contain the following tag columns: ' + ', '.join(
+        return 'Integration.upsert_sites(): data_frame must contain the following tag columns: ' + ', '.join(
             tag_columns)
     elif tag_columns is None:
         tag_columns = []
@@ -491,7 +493,6 @@ def upsert_sites(df: pandas.DataFrame, api_inputs: ApiInputs, tag_columns: list
         return response_status, pandas.DataFrame()
 
     if save_additional_columns_as_slices and slices_data_frame.shape[0] > 0:
-        # print('Doing Slices: ' + str(save_additional_columns_as_slices) + ' - ' + str(slices_data_frame.shape[0]))
         slices_merged = pandas.merge(left=response_data_frame, right=slices_data_frame, left_on='installationCode',
                                      right_on='InstallationCode')
         slices_data_frame = slices_merged[['installationId', 'Slices']]
@@ -641,8 +642,7 @@ def upsert_workorders(df: pandas.DataFrame, api_inputs: ApiInputs, save_addition
                                     'NonBillableCost', 'Location', 'RawLocation', 'ScheduledStartDate',
                                     'ScheduledCompletionDate', 'Meta']]
 
-    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name='WorkOrder',
-                                api_project_id=api_inputs.api_project_id, is_timeseries=True)
+    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name='WorkOrder')
     json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "append",
                     "tableDef": _get_structure(data_frame),
                     "keyColumns": ["WorkOrderId"]
@@ -762,7 +762,7 @@ def upsert_timeseries_ds(df: pandas.DataFrame, api_inputs: ApiInputs, is_local_t
 
     def in_range(row):
         j_row = (row['Timestamp'] >= row['dateFrom']) & (
-                row['Timestamp'] < (row['dateTo'] + datetime.timedelta(days=1)))
+            row['Timestamp'] < (row['dateTo'] + datetime.timedelta(days=1)))
         return str(j_row)
 
     data_frame['InDateRange'] = data_frame.apply(in_range, axis=1)
@@ -792,10 +792,10 @@ def upsert_timeseries_ds(df: pandas.DataFrame, api_inputs: ApiInputs, is_local_t
         if row[date_col].minute < 15:
             j_row = row[date_col].replace(minute=0)
             return j_row
-        elif row[date_col].minute >= 15 & row[date_col].minute < 29:
+        elif row[date_col].minute >= 15 and row[date_col].minute < 30:
             j_row = row[date_col].replace(minute=15)
             return j_row
-        elif row[date_col].minute >= 30 & row[date_col].minute < 44:
+        elif row[date_col].minute >= 30 and row[date_col].minute < 45:
             j_row = row[date_col].replace(minute=30)
             return j_row
         else:
@@ -813,12 +813,9 @@ def upsert_timeseries_ds(df: pandas.DataFrame, api_inputs: ApiInputs, is_local_t
 
     data_frame = data_frame.loc[:, ['ObjectPropertyId', 'Timestamp', 'TimestampId', 'TimestampLocal',
                                     'TimestampLocalId', 'Value', 'DataFeedFileStatusId', 'InstallationId', 'Meta']]
-
     name = 'Timeseries_Ds'
 
-    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=name,
-                                api_project_id=api_inputs.api_project_id, is_timeseries=True,
-                                batch_id=data_feed_file_status_id)
+    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=name, batch_id=data_feed_file_status_id)
 
     json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "upsert",
                     "isLocalTime": is_local_time,
@@ -856,6 +853,213 @@ def upsert_timeseries_ds(df: pandas.DataFrame, api_inputs: ApiInputs, is_local_t
     return response_status, response_df
 
 
+@_with_func_attrs(df_required_columns=['ObjectPropertyId', 'InstallationId', 'Timestamp', 'Value'])
+def upsert_timeseries(df: pandas.DataFrame, api_inputs: ApiInputs, is_local_time: bool = True,
+                      save_additional_columns_as_slices: bool = False, data_feed_file_status_id: uuid.UUID = None):
+    """Upserts timeseries to EventHub for processing.
+
+    The following columns are required to be present in the data_frame:
+    - InstallationId
+    - ObjectPropertyId
+    - Timestamp
+    - Value
+
+    Parameters
+    ----------
+    df: pandas.DataFrame
+        Dataframe containing the data to be appended to timeseries.
+    api_inputs: ApiInputs
+        Object returned by initialize() function.
+    is_local_time : bool, default = True
+         Whether the datetime values are in local time or UTC. If false, then UTC (Default value = True).
+    save_additional_columns_as_slices : bool, default = False
+         (Default value = False)
+    data_feed_file_status_id : uuid.UUID, default = None
+         Enables developer to identify upserted rows using during development. This data is posted to the
+         DataFeedFileStatusId in the Timeseries_Ds table.
+
+         Once deployed, the DataFeedFileStatusId field will contain a unique Guid which will assist in
+         tracking upload results and logging.
+
+    Returns
+    -------
+    tuple[str, pandas.DataFrame]
+        (response_status, response_df) - Returns the response status and the dataframe containing the parsed response
+        text.
+
+    """
+
+    data_frame = df.copy()
+
+    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
+        logger.error("You must call initialize() before using API.")
+        return pandas.DataFrame()
+
+    required_columns = ['ObjectPropertyId', 'InstallationId', 'Timestamp', 'Value']
+    req_cols = ', '.join(required_columns)
+    proposed_columns = data_frame.columns.tolist()
+
+    if not set(required_columns).issubset(proposed_columns):
+        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
+        return 'integration.upsert_timeseries_ds() - data_frame must contain the following columns: ' + req_cols
+
+    slice_columns = set(proposed_columns).difference(set(required_columns))
+
+    if len(list(slice_columns)) > 0 and save_additional_columns_as_slices is True:
+        def update_values(row):
+            j_row = row[list(slice_columns)].to_json()
+            return str(j_row)
+
+        data_frame['Meta'] = data_frame[list(slice_columns)].assign(
+            **data_frame[list(slice_columns)].select_dtypes(['datetime', 'object']).astype(str)).apply(update_values, axis=1)
+
+        data_frame = data_frame.drop(columns=list(slice_columns))
+    elif len(list(slice_columns)) > 0 and save_additional_columns_as_slices is not True:
+        data_frame = data_frame.drop(columns=list(slice_columns))
+        data_frame['Meta'] = ''
+    else:
+        data_frame['Meta'] = ''
+
+    if api_inputs.data_feed_file_status_id is not None and \
+            api_inputs.data_feed_file_status_id != '00000000-0000-0000-0000-000000000000':
+        data_frame['DataFeedFileStatusId'] = api_inputs.data_feed_file_status_id
+    elif data_feed_file_status_id is not None:
+        data_frame['DataFeedFileStatusId'] = data_feed_file_status_id
+    else:
+        data_frame['DataFeedFileStatusId'] = '00000000-0000-0000-0000-000000000000'
+
+    if not pandas.api.types.is_datetime64_ns_dtype(data_frame['Timestamp']):
+        logger.exception("Timestamp series value is not of type: datetime64[ns] dtype.")
+        return "Timestamp series value is not of type: datetime64[ns] dtype."
+
+    site_list = data_frame['InstallationId'].unique().tolist()
+    start_date = data_frame['Timestamp'].min(axis=0, skipna=True)
+    end_date = data_frame['Timestamp'].max(axis=0, skipna=True)
+
+    timezones_df = _timezone_dst_offsets(api_inputs=api_inputs, date_from=start_date.date(), date_to=end_date.date(),
+                                         installation_id_list=site_list)
+
+    if timezones_df.empty:
+        logger.exception('Timezone DST offsets failed to retrieve.')
+        return 'Timezone DST offsets failed to retrieve.'
+
+    def convert_dst_interval_dates(row):
+        row['start'] = pd.to_datetime(row['start'])
+        row['end'] = pd.to_datetime(row['end'])
+        row['offsetToUtcMinutes'] = row['standardOffsetUtc'] + row['dstOffsetUtc']
+        return row
+
+    timezones_df = timezones_df.apply(convert_dst_interval_dates, axis=1)
+
+    if not is_local_time:
+        def convert_dst_interval_range_to_utc(row):
+            dst_start = row['start']
+            dst_end = row['end']
+
+            if row['standardOffsetUtc'] >= 0:
+                dst_start = dst_start - (datetime.timedelta(minutes=row['standardOffsetUtc'])) - datetime.timedelta(minutes=row['dstOffsetUtc'])
+                dst_end = dst_end - (datetime.timedelta(minutes=row['standardOffsetUtc'])) - datetime.timedelta(minutes=row['dstOffsetUtc'])
+            else:
+                dst_start = dst_start + (datetime.timedelta(minutes=row['standardOffsetUtc'])) + datetime.timedelta(minutes=row['dstOffsetUtc'])
+                dst_end = dst_end + (datetime.timedelta(minutes=row['standardOffsetUtc'])) + datetime.timedelta(minutes=row['dstOffsetUtc'])
+            row['start'] = dst_start
+            row['end'] = dst_end
+
+            return row
+
+        timezones_df = timezones_df.apply(convert_dst_interval_range_to_utc, axis=1)
+        timezones_df = timezones_df.drop(columns=['timezoneId', 'timezoneName', 'standardOffsetUtc', 'dstOffsetUtc'])
+
+    data_frame = data_frame.merge(timezones_df, left_on='InstallationId', right_on='installationId', how='inner')
+    data_frame = data_frame[(data_frame.Timestamp >= data_frame.start) & (data_frame.Timestamp <= data_frame.end) & (data_frame.Timestamp.apply(lambda x: x.year) == data_frame.year)]
+
+    if is_local_time:
+        def to_utc(row):
+            if row['offsetToUtcMinutes'] >= 0:
+                j_row = row['TimestampLocal'] - datetime.timedelta(minutes=row['offsetToUtcMinutes'])
+            else:
+                j_row = row['TimestampLocal'] + datetime.timedelta(minutes=abs(row['offsetToUtcMinutes']))
+            return j_row
+
+        data_frame = data_frame.assign(TimestampLocal=data_frame['Timestamp'])
+        data_frame['Timestamp'] = data_frame.apply(to_utc, axis=1)
+
+    elif not is_local_time:
+        def from_utc(row):
+            if row['offsetToUtcMinutes'] >= 0:
+                j_row = row['Timestamp'] + datetime.timedelta(minutes=row['offsetToUtcMinutes'])
+            else:
+                j_row = row['Timestamp'] - datetime.timedelta(minutes=abs(row['offsetToUtcMinutes']))
+            return j_row
+        data_frame['TimestampLocal'] = data_frame.apply(from_utc, axis=1)
+
+    def bin_to_15_minute_interval(row, date_col):
+        if row[date_col].minute < 15:
+            j_row = row[date_col].replace(minute=0)
+            return j_row
+        elif row[date_col].minute >= 15 and row[date_col].minute < 30:
+            j_row = row[date_col].replace(minute=15)
+            return j_row
+        elif row[date_col].minute >= 30 and row[date_col].minute < 45:
+            j_row = row[date_col].replace(minute=30)
+            return j_row
+        else:
+            j_row = row[date_col].replace(minute=45)
+            return j_row
+
+    def convert_to_iso_format_string(row):
+        timestamp_format = "%Y-%m-%dT%H:%M:%SZ"
+        row['Timestamp'] = row['Timestamp'].to_pydatetime().replace(second=0, microsecond=0).strftime(timestamp_format)
+        row['TimestampLocal'] = row['TimestampLocal'].to_pydatetime().replace(second=0, microsecond=0).strftime(timestamp_format)
+        row['TimestampId'] = row['TimestampId'].to_pydatetime().replace(second=0, microsecond=0).strftime(timestamp_format)
+        row['TimestampLocalId'] = row['TimestampLocalId'].to_pydatetime().replace(second=0, microsecond=0).strftime(timestamp_format)
+        return row
+
+    data_frame['TimestampId'] = data_frame.apply(bin_to_15_minute_interval, args=('Timestamp',), axis=1)
+    data_frame['TimestampLocalId'] = data_frame.apply(bin_to_15_minute_interval, args=('TimestampLocal',), axis=1)
+    data_frame = data_frame.apply(convert_to_iso_format_string, axis=1)
+
+    headers = api_inputs.api_headers.integration
+
+    logger.info("Upserting data to Timeseries.")
+
+    data_frame = data_frame.loc[:, ['ObjectPropertyId', 'Timestamp', 'TimestampId', 'TimestampLocal',
+                                    'TimestampLocalId', 'Value', 'DataFeedFileStatusId', 'InstallationId', 'Meta']]
+
+    container = 'data-ingestion-timeseries-adx'
+    folder = 'to-eventhub'
+    name = 'Timeseries'
+
+    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, container=container, folder=folder, name=name, batch_id=data_feed_file_status_id, include_header=True)
+
+    json_payload = {"path": upload_result[0], "fileCount": upload_result[1]}
+
+    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/timeseries"
+    logger.info("Sending request: POST %s", url)
+
+    logger.info("Sending request to ingest %s files from %s", str(upload_result[1]), upload_result[0])
+
+    response = requests.post(url, json=json_payload, headers=headers)
+    response_status = '{} {}'.format(response.status_code, response.reason)
+
+    if response.status_code != 200:
+        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
+                     response.reason)
+        return response_status, pandas.DataFrame()
+    elif len(response.text) == 0:
+        logger.error('No data returned for this API call. %s', response.request.url)
+        return response_status, pandas.DataFrame()
+    elif response.status_code == 200:
+        _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=data_frame.shape[0])
+
+    response_df = pandas.read_json(response.text, typ='series').to_frame().T
+    response_df.columns = _column_name_cap(columns=response_df.columns)
+
+    logger.info("Response status: %s", response_status)
+    logger.info("Ingestion complete. ")
+    return response_status, response_df
+
+
 def upsert_data(data_frame, api_inputs: ApiInputs, table_name: str, key_columns: list,
                 is_slices_table=False):
     """Upsert data
@@ -888,7 +1092,6 @@ def upsert_data(data_frame, api_inputs: ApiInputs, table_name: str, key_columns:
         logger.error("You must call initialize() before using API.")
         return pandas.DataFrame()
 
-    # payload = {}
     headers = api_inputs.api_headers.integration
 
     if len(key_columns) == 0 or key_columns is None:
@@ -903,8 +1106,7 @@ def upsert_data(data_frame, api_inputs: ApiInputs, table_name: str, key_columns:
         table_structure['Slices'] = 'dynamic'
 
     # upload Blobs to folder
-    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name,
-                                api_project_id=api_inputs.api_project_id, is_timeseries=False)
+    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name)
     json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "upsert",
                     "tableDef": table_structure, "keyColumns": key_columns}
 
@@ -959,14 +1161,12 @@ def replace_data(data_frame, api_inputs: ApiInputs, table_name: str):
         logger.error("You must call initialize() before using API.")
         return pandas.DataFrame()
 
-    # payload = {}
     headers = api_inputs.api_headers.integration
 
     logger.info("Replacing all data for %s", table_name)
 
     # upload Blobs to folder
-    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name,
-                                api_project_id=api_inputs.api_project_id, is_timeseries=False)
+    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name)
     json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "replace",
                     "tableDef": _get_structure(data_frame)}
     logger.info(json_payload)
@@ -1028,8 +1228,7 @@ def append_data(data_frame, api_inputs: ApiInputs, table_name: str):
     logger.info("Appending data for %s", table_name)
 
     # upload Blobs to folder
-    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name,
-                                api_project_id=api_inputs.api_project_id, is_timeseries=False)
+    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name)
     json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "append",
                     "tableDef": _get_structure(data_frame)}
 
@@ -1133,8 +1332,8 @@ def upsert_event_work_order_id(api_inputs: ApiInputs, event_task_id: uuid.UUID,
     header = api_inputs.api_headers.default
 
     payload = {
-                  "IntegrationId": integration_id,
-                  "Status": work_order_status
+        "IntegrationId": integration_id,
+        "Status": work_order_status
     }
 
     url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/events/{str(event_task_id)}/work-order"
@@ -1147,4 +1346,3 @@ def upsert_event_work_order_id(api_inputs: ApiInputs, event_task_id: uuid.UUID,
         return response_status, response.text
 
     return response_status, response.text
-
